//
// AppPreferences - manage key=value pairs in persistent store
//

import java.util.Vector;
import javax.microedition.rms.*;

/**
 * AppPreferences store application preferences as key=value pairs in a RecordStore
 * associated with the midlet.
 *
 * @see javax.microedition.rms.RecordStore
 *
 * @author Edward Hutchins
 * Version: 2.0M  Copyright (C) 2006 Hutchins Software
 * portions Copyright (C) 2000 Russell Marks
 * portions Copyright (C) 1996 Rasmus Ekman
 * Version: 1.1  Copyright (C) 1990,91 Hutchins Software
 * This software is licenced under the GNU General Public Licence
 * Please read the associated legal documentation
 */

public class AppPreferences
{
    private String name = null; // RecordStore name
    private Vector data = null; // set of PrefRecords parsed from RecordStore
    private boolean dirty = false; // if true, changes are pending

    /**
     * PrefRecord keeps track of the RecordStore id, key and value
     * (stored in the AppPreference's data Vector)
     */

    private class PrefRecord
    {
	private int id; // record ID
	private String key;
	private String val;

	/**
	 * creates a new record with the specified RecordStore id, key and value
	 * @param id RecordStore identifier
	 * @param key key name
	 * @param val value
	 */
	public PrefRecord( int id, String key, String val )
	{
	    this.id = id;
	    this.key = key;
	    this.val = val;
	}
	
	/**
	 * generates a cannonical "key=value" byte sequence
	 * @return the byte array corresponding to "key=value"
	 */
	public byte[] getRecord()
	{
	    byte[] k = key.getBytes();
	    byte[] v = val.getBytes();
	    byte[] buf = new byte[k.length + 1 + v.length];
	    int i = 0;
	    for (int j = 0; j < k.length; ++j)
		buf[i++] = k[j];
	    buf[i++] = '=';
	    for (int j = 0; j < v.length; ++j) 
		buf[i++] = v[j];
	    return( buf );
	}
    }

    /**
     * This method must be called in order to flush in-memory changes to the
     * preferences into the underlying RecordStore.
     */

    public void sync()
    {
	if (!dirty) return;
	HyperoidMidlet.conPrint( "Syncing record store " + name );
	try
	{
	    RecordStore.deleteRecordStore( name );
	}
	catch (Exception e)
	{
	    HyperoidMidlet.conPrint( "Sync delete failed (" + e.toString() + ")" );
	}
	try
	{
	    RecordStore rs = RecordStore.openRecordStore( name, true );
	    for (int i = 0; i < data.size(); ++i)
	    {
		PrefRecord r = (PrefRecord)data.elementAt( i );
		byte[] kv = r.getRecord();
		r.id = rs.addRecord( kv, 0, kv.length );
		HyperoidMidlet.conPrint( r.id + ": " + new String( kv ) );
	    }
	    rs.closeRecordStore();
	    HyperoidMidlet.conPrint( "Done syncing record store " + name );
	    dirty = false;
	}
	catch (Exception e)
	{
	    HyperoidMidlet.conPrint( "Failed to sync record store (" + e.toString() + ")" );
	}
    }

    /**
     * This method empties the set of key=value pairs and the underlying RecordStore,
     * enabling future calls to get() to return the default values.
     */

    public void reset()
    {
	try
	{
	    HyperoidMidlet.conPrint( "Resetting record store " + name );
	    RecordStore.deleteRecordStore( name );
	    RecordStore rs = RecordStore.openRecordStore( name, true );
	    rs.closeRecordStore();
	    data.removeAllElements();
	}
	catch (Exception e)
	{
	    HyperoidMidlet.conPrint( "Failed to reset record store (" + e.toString() + ")" );
	}
    }

    /**
     * Takes the string name for the underlying RecordStore. The RecordStore
     * is only read once at object construction, all further interactions are
     * assumed to happen via the AppPreferences object (i.e. sync()'ing will
     * obliterate any subsequent changes to the RecordStore.
     *
     * @param name The 32-character name of the underlying RecordStore.
     */

    public AppPreferences( String name )
    {
	this.name = name;
	data = new Vector();
	HyperoidMidlet.conPrint( "Reading record store " + name );
	try
	{
	    RecordStore rs = RecordStore.openRecordStore( name, true );
	    RecordEnumeration cursor = rs.enumerateRecords( null, null, false );
	    while (cursor.hasNextElement())
	    {
		int id = cursor.nextRecordId();
		byte[] buf = rs.getRecord( id );
		HyperoidMidlet.conPrint( id + "? " + new String( buf ) );
		for (int j = 0; j < buf.length; ++j)
		{
		    if (buf[j] == '=')
		    {
			PrefRecord p = new PrefRecord( id, new String( buf, 0, j ), new String( buf, j + 1, buf.length - (j + 1) ) );
			// avoid duplicate keys in the (possibly corrupt) record store
			int idx = find( p.key );
			if (idx >= 0)
			{
			    data.setElementAt( p, idx );
			    PrefRecord r = (PrefRecord)data.elementAt( idx );
			    HyperoidMidlet.conPrint( "Duplicate key " + r.id + ": " + r.key + "=" + r.val );
			}
			else
			{
			    data.addElement( p );
			    HyperoidMidlet.conPrint( id + ": " + p.key + "=" + p.val );
			}
		    }
		}
	    }
	    cursor.destroy();
	    rs.closeRecordStore();
	    HyperoidMidlet.conPrint( "Done reading record store " + name );
	}
	catch (Exception e)
	{
	    HyperoidMidlet.conPrint( "Failed to access record store (" + e.toString() + ")" );
	}
    }

    /**
     * find returns an internal index value to the specified key,
     * which matches case-insensitively. If the key isn't found it
     * returns -1 and does not create a new empty key= pair.
     *
     * @param key The key to find
     * @return A positive value if the key exists, or -1 if it was not found.
     */

    public int find( String key )
    {
	for (int i = 0; i < data.size(); ++i)
	{
	    PrefRecord r = (PrefRecord)data.elementAt( i );
	    if (r.key.equalsIgnoreCase( key ))
		return( i );
	}
	return( -1 );
    }

    /**
     * set sets the key's value, creating the key=value pair if it doesn't already
     * exist.
     *
     * @param key The key to set.
     * @param val The value to set key to.
     * 
     * @see #sync()
     */

    public void set( String key, String val )
    {
	int i = find( key );
	if (i >= 0)
	{
	    PrefRecord r = (PrefRecord)data.elementAt( i );
	    r.val = val;
	    HyperoidMidlet.conPrint( "Set " + r.id + ": " + r.key + "=" + r.val );
	}
	else
	{
	    PrefRecord p = new PrefRecord( -1, key, val );
	    data.addElement( p );
	    HyperoidMidlet.conPrint( "Set new: " + p.key + "=" + p.val );
	}
	dirty = true;
    }
    
    /**
     * get looks up the key in the set of preferences, and returns the key's value
     * if found, or the default_val if not. If the key is not present it will be
     * added along with the default_val.
     * 
     * @param key Key to look up.
     * 
     * @param default_val Value to return if key is not present.
     * @return Either the found value, or the default value if a key doesn't match.
     */

    public String get( String key, String default_val )
    {
	int i = find( key );
	if (i >= 0)
	{
	    PrefRecord r = (PrefRecord)data.elementAt( i );
	    HyperoidMidlet.conPrint( "Found " + r.id + ": " + r.key + "=" + r.val );
	    return( r.val );
	}
	set( key, default_val );
	return( default_val );
    }
}
